@@ -3,6 +3,7 @@ from __future__ import absolute_import, print_function
 
 import abc
 import os
+import sys
 import traceback
 import uuid
 
@@ -23,7 +24,7 @@ from suanpan.stream import handlers, loops
 from suanpan.stream.interfaces import HasHandlers
 from suanpan.stream.objects import Context
 from suanpan.stream.sio import StreamServer, WebSocketHandler
-from suanpan.utils import functional, json
+from suanpan.utils import functional, json, port
 
 
 class Stream(
@@ -53,14 +54,16 @@ class Stream(
         self.argsDict = {}
         self.options = None
         self.args = None
+        self._exit_code = 0
 
         self.callLoop = StreamCallLoop(self)
         self.triggerLoop = StreamTriggerLoop(self)
         self.sioLoop = StreamSIOLoop(self)
+        self.httpLoop = StreamHTTPLoop(self)
 
     @property
     def loops(self):
-        return [self.callLoop, self.triggerLoop, self.sioLoop]
+        return [self.callLoop, self.triggerLoop, self.sioLoop, self.httpLoop]
 
     def beforeInit(self):
         logger.logDebugInfo()
@@ -150,6 +153,15 @@ class Stream(
         runners = (loop.start for loop in self.loops if loop.ready)
         asyncio.wait(asyncio.run(runners))
 
+        if self._exit_code != 0:
+            sys.exit(self._exit_code)
+
+    def close(self, code=0):
+        self._exit_code = code
+        for loop in self.loops:
+            if loop.ready:
+                loop.close()
+
     @property
     def vars(self):
         return mstorage.vars
@@ -265,7 +277,7 @@ class StreamMQBaseLoop(StreamBaseLoop):
                 self.sendSuccessMessage(message, outputs)
         except Exception as e:  # pylint: disable=broad-except
             tracebackInfo = traceback.format_exc()
-            logger.error(tracebackInfo)
+            logger.error(tracebackInfo, extra={"success": "false"})
             self.sendFailureMessage(message, tracebackInfo)
             self.stream.sioLoop.notifyError(e)
 
@@ -312,6 +324,10 @@ class StreamCallLoop(StreamMQBaseLoop):
             queue=self.stream.options["stream-recv-queue"],
         )
 
+    def close(self):
+        logger.debug("Stream Call Loop Closing...")
+        mq.close()
+
     @runtime.globalrun
     def start(self):
         logger.debug("Stream Call Loop Starting...")
@@ -341,6 +357,10 @@ class StreamTriggerLoop(StreamMQBaseLoop):
     def loop(self, funcOrIter, *_, **__):
         self._loop = funcOrIter if callable(funcOrIter) else lambda: funcOrIter
 
+    def close(self):
+        logger.debug("Stream Trigger Loop Closing...")
+        self._loop.close()
+
     @runtime.globalrun
     def start(self):
         logger.debug("Stream Trigger Loop Starting...")
@@ -360,6 +380,7 @@ class StreamSIOLoop(StreamBaseLoop):
         self.extraStaticFiles = {}
         self.staticFiles = {}
         self._enabled = True
+        self.server = None
 
     def _getDefaultOptions(self):
         options = Context()
@@ -446,11 +467,112 @@ class StreamSIOLoop(StreamBaseLoop):
             "traceback": traceback.format_exc(),
         }
 
+    def close(self):
+        logger.debug("Stream SIO Loop Closing...")
+        if self.server:
+            self.server.close()
+
+    def _init_win_server(self):
+        p = None
+        for _ in range(20):
+            try:
+                p = port.get_free_port()
+                self.server.set_listener(("", p))
+                self.server.start()
+                logger.info("Stream SIO Loop listen port {}".format(p))
+                port.register_server(g.port, p)
+                return
+            except OSError as e:
+                if e.errno in [10013, 10048]:
+                    logger.warn("Stream SIO Loop listen port conflict {}".format(p))
+                    continue
+
+                raise
+            except Exception:
+                raise
+
+        raise Exception("port conflicts repeatedly failed, please try again later")
+
+    def _start_server(self):
+        if port.need_free_port():
+            self._init_win_server()
+        else:
+            self.server.start()
+
     @runtime.globalrun
     def start(self):
         logger.debug("Stream SIO Loop Starting...")
         self.staticFiles = self.getDefaultStaticFiles(self.staticFolder)
         app = socketio.WSGIApp(self.sio, static_files=self.staticFiles)
-        gevent.pywsgi.WSGIServer(
+        self.server = gevent.pywsgi.WSGIServer(
             ("", g.port), app, handler_class=WebSocketHandler
-        ).serve_forever()
+        )
+        self._start_server()
+        self.server.serve_forever()
+
+
+class StreamHTTPLoop(StreamBaseLoop):
+    def __init__(self, stream):
+        super().__init__(stream)
+        self._enabled = True
+        self.server = None
+
+    @property
+    def ready(self):
+        return self._enabled or g.debug
+
+    def enable(self):
+        self._enabled = True
+        return self
+
+    def disable(self):
+        self._enabled = False
+        return self
+
+    def application(self, env, start_response):
+        if env['REQUEST_METHOD'] == 'POST' and env['PATH_INFO'] == '/internal/trap':
+            start_response('200 OK', [('Content-Type', 'text/json')])
+            self.stream.close(code=98)
+            return [json.dumps({"message": "ok"}, ensure_ascii=False).encode('utf8')]
+
+        start_response('404 Not Found', [('Content-Type', 'text/json')])
+        return [b'{"message": "not found"}']
+
+    def close(self):
+        logger.debug("Stream HTTP Loop Closing...")
+        if self.server:
+            self.server.close()
+
+    def _init_win_server(self):
+        p = None
+        for _ in range(20):
+            try:
+                p = port.get_free_port()
+                self.server.set_listener(("", p))
+                self.server.start()
+                logger.info("Stream HTTP Loop listen port {}".format(p))
+                port.register_server(g.termPort, p)
+                return
+            except OSError as e:
+                if e.errno in [10013, 10048]:
+                    logger.warn("Stream HTTP Loop listen port conflict {}".format(p))
+                    continue
+
+                raise
+            except Exception:
+                raise
+
+        raise Exception("port conflicts repeatedly failed, please try again later")
+
+    def _start_server(self):
+        if port.need_free_port():
+            self._init_win_server()
+        else:
+            self.server.start()
+
+    @runtime.globalrun
+    def start(self):
+        logger.debug("Stream HTTP Loop Starting...")
+        self.server = gevent.pywsgi.WSGIServer(("", g.termPort), self.application)
+        self._start_server()
+        self.server.serve_forever()
